@@ -157,15 +157,23 @@ print_message_id (smtp_message_t message, struct rfc2822_header *header)
 {
   const char *message_id;
   char buf[64];
-  static int generation;
+#ifdef HAVE_GETTIMEOFDAY
+  struct timeval tv;
+#endif
 
   assert (message != NULL && header != NULL);
 
   message_id = header->value;
   if (message_id == NULL)
     {
-      snprintf (buf, sizeof buf, "%ld.%d@%s",
-		time (NULL), generation++, message->session->localhost);
+#ifdef HAVE_GETTIMEOFDAY
+      if (gettimeofday (&tv, NULL) == -1) /* This shouldn't fail ... */
+	snprintf (buf, sizeof buf, "%ld.%ld.%d@%s", tv.tv_sec, tv.tv_usec,
+		  getpid (), message->session->localhost);
+      else /* ... but if it does fall back to using time() */
+#endif
+      snprintf (buf, sizeof buf, "%ld.%d@%s", time (NULL),
+      	        getpid (), message->session->localhost);
       message_id = buf;
     }
   /* TODO: implement line folding at white spaces */
