@@ -100,9 +100,9 @@ next_recipient (smtp_recipient_t recipient)
   return recipient;
 }
 
-/* Return a pointer to the next unsent message.
+/* Set the session's current message to the next unsent message.
  */
-static int
+int
 next_message (smtp_session_t session)
 {
   while ((session->current_message = session->current_message->next) != NULL)
@@ -590,7 +590,7 @@ void
 cmd_greeting (siobuf_t conn, smtp_session_t session)
 {
   /* Set a five minute timeout. */
-  sio_set_timeout (conn, 5 * 60 * 1000);
+  sio_set_timeout (conn, session->greeting_timeout);
   session->cmd_state = -1;
 }
 
@@ -663,14 +663,30 @@ report_extensions (smtp_session_t session)
       if (quit_now)
         exts |= EXT_DSN;
     }
+#ifdef USE_CHUNKING
+  if (no_required_extension (session, EXT_CHUNKING))
+    {
+      quit_now = 0;
+      if (session->event_cb != NULL)
+	(*session->event_cb) (session, SMTP_EV_EXTNA_CHUNKING,
+			      session->event_cb_arg, &quit_now);
+      if (quit_now)
+        exts |= EXT_CHUNKING;
+    }
+  if (no_required_extension (session, EXT_BINARYMIME))
+    {
+      if (session->event_cb != NULL)
+	(*session->event_cb) (session, SMTP_EV_EXTNA_BINARYMIME,
+			      session->event_cb_arg);
+      exts |= EXT_BINARYMIME;
+    }
+#endif
   if (no_required_extension (session, EXT_8BITMIME))
     {
-      quit_now = 1;
       if (session->event_cb != NULL)
 	(*session->event_cb) (session, SMTP_EV_EXTNA_8BITMIME,
 			      session->event_cb_arg);
-      if (quit_now)
-        exts |= EXT_8BITMIME;
+      exts |= EXT_8BITMIME;
     }
 #ifdef USE_ETRN
   if (no_required_extension (session, EXT_ETRN))
@@ -723,7 +739,7 @@ cb_ehlo (smtp_session_t session, char *buf)
       set_auth_mechanisms (session, p);
     }
 #endif
-  else if (strcasecmp (token, "STARTTLS") == 0)		/* RFC 2487 */
+  else if (strcasecmp (token, "STARTTLS") == 0)		/* RFC 3207 */
     session->extensions |= EXT_STARTTLS;
   else if (strcasecmp (token, "SIZE") == 0)		/* RFC 1870 */
     {
@@ -743,10 +759,10 @@ cb_ehlo (smtp_session_t session, char *buf)
     }
   else if (strcasecmp (token, "ETRN") == 0)		/* RFC 1985 */
     session->extensions |= EXT_ETRN;
-#ifdef USE_XUSR
   else if (strcasecmp (token, "XUSR") == 0)	/* sendmail (I feel ill) */
     session->extensions |= EXT_XUSR;
-#endif
+  else if (strcasecmp (token, "XEXCH50") == 0)	/* exchange (I feel worse) */
+    session->extensions |= EXT_XEXCH50;
   return 1;
 }
 
@@ -759,9 +775,8 @@ rsp_ehlo (siobuf_t conn, smtp_session_t session)
   session->extensions = 0;
   destroy_auth_mechanisms (session);
   code = read_smtp_response (conn, session, &session->mta_status, cb_ehlo);
-  if (code == 0)
+  if (code < 0)
     {
-      set_error (SMTP_ERR_INVALID_RESPONSE_SYNTAX);
       session->rsp_state = S_quit;
       return;
     }
@@ -879,13 +894,15 @@ void
 rsp_helo (siobuf_t conn, smtp_session_t session)
 {
   int code;
+#ifdef USE_TLS
+  int notls;
+#endif
 
   session->extensions = 0;
   destroy_auth_mechanisms (session);
   code = read_smtp_response (conn, session, &session->mta_status, NULL);
-  if (code == 0)
+  if (code < 0)
     {
-      set_error (SMTP_ERR_INVALID_RESPONSE_SYNTAX);
       session->try_fallback_server = 1;
       session->rsp_state = S_quit;
       return;
@@ -899,9 +916,22 @@ rsp_helo (siobuf_t conn, smtp_session_t session)
       return;
     }
 
-  /* There are no extensions.  Make sure none were required.
-   */
-  if (!report_extensions (session))
+#ifdef USE_TLS
+  /* Care must be taken when checking for required TLS.  The client may
+     have connected to a tunnel which offers the STARTTLS extension and
+     the real server does not implement SMTP extensions.  Hence the
+     check that TLS is actually in use before reporting that the
+     extension is not available.  */
+  notls = !session->using_tls && session->starttls_enabled == Starttls_REQUIRED;
+  if (notls && session->event_cb != NULL)
+    (*session->event_cb) (session, SMTP_EV_EXTNA_STARTTLS,
+			  session->event_cb_arg, NULL);
+#else
+# define notls 0
+#endif
+
+  /* There are no extensions.  Make sure none were required. */
+  if (!report_extensions (session) || notls)
     {
       set_error (SMTP_ERR_EXTENSION_NOT_AVAILABLE);
       session->rsp_state = S_quit;
@@ -933,7 +963,7 @@ cmd_mail (siobuf_t conn, smtp_session_t session)
 
   /* Set a five minute timeout.  This stays in force until the DATA
      command. */
-  sio_set_timeout (conn, 5 * 60 * 1000);
+  sio_set_timeout (conn, session->envelope_timeout);
 
   message = session->current_message;
   mailbox = message->reverse_path_mailbox;
@@ -956,8 +986,12 @@ cmd_mail (siobuf_t conn, smtp_session_t session)
 		    encode_xtext (xtext, sizeof xtext, message->dsn_envid));
     }
 
-  /* 8BITMIME: RET=FULL/HDRS  ENVID=xtext */
+  /* 8BITMIME: BODY=7BIT/8BITMIME/BINARYMIME */
+#ifdef USE_CHUNKING
+  if ((session->extensions & (EXT_8BITMIME | EXT_BINARYMIME))
+#else
   if ((session->extensions & EXT_8BITMIME)
+#endif
       && message->e8bitmime != E8bitmime_NOTSET)
     {
       sio_write (conn, " BODY=", -1);
@@ -965,6 +999,10 @@ cmd_mail (siobuf_t conn, smtp_session_t session)
 	sio_write (conn, "8BITMIME", -1);
       else if (message->e8bitmime == E8bitmime_7BIT)
 	sio_write (conn, "7BIT", -1);
+#ifdef USE_CHUNKING
+      else if (message->e8bitmime == E8bitmime_BINARYMIME)
+	sio_write (conn, "BINARYMIME", -1);
+#endif
     }
 
   if ((session->extensions & EXT_DELIVERBY) && message->by_mode != By_NOTSET)
@@ -1012,6 +1050,11 @@ rsp_mail (siobuf_t conn, smtp_session_t session)
   message = session->current_message;
   code = read_smtp_response (conn, session,
   		             &message->reverse_path_status, NULL);
+  if (code < 0)
+    {
+      session->rsp_state = S_quit;
+      return;
+    }
 
   /* Notify the MAIL FROM: status */
   if (session->event_cb != NULL)
@@ -1027,9 +1070,7 @@ rsp_mail (siobuf_t conn, smtp_session_t session)
   else
     {
       message->valid_recipients = 0;
-#ifdef USE_REQUIRE_ALL_RECIPIENTS
       message->failed_recipients = 0;
-#endif
       session->rsp_state = S_rcpt;
     }
 }
@@ -1095,15 +1136,11 @@ cmd_rcpt (siobuf_t conn, smtp_session_t session)
   session->cmd_recipient = next_recipient (session->cmd_recipient);
   if (session->cmd_recipient != NULL)
     session->cmd_state = S_rcpt;
-  else
-#ifdef USE_REQUIRE_ALL_RECIPIENTS
-  /* Ugly hack:  if requiring all recpients to succeed the decision can't
-                 be made here.  So PIPELINING's improvement is lost. */
-  if (session->require_all_recipients)
+  else if (session->require_all_recipients)
+    /* can't pipeline the DATA command when require_all_recpients is set. */
     session->cmd_state = -1;
   else
-#endif
-#if 0
+#ifdef USE_CHUNKING
     session->cmd_state = (session->extensions & EXT_CHUNKING) ? S_bdat : S_data;
 #else
     session->cmd_state = S_data;
@@ -1117,12 +1154,16 @@ rsp_rcpt (siobuf_t conn, smtp_session_t session)
 
   code = read_smtp_response (conn, session,
   			     &session->rsp_recipient->status, NULL);
+  if (code < 0)
+    {
+      session->rsp_state = S_quit;
+      return;
+    }
+
   if (code == 2)
     session->current_message->valid_recipients += 1;
-#ifdef USE_REQUIRE_ALL_RECIPIENTS
   else
     session->current_message->failed_recipients += 1;
-#endif
 
   /* MTA will never accept this recipient.  Make sure it isn't used
      again. */
@@ -1138,21 +1179,14 @@ rsp_rcpt (siobuf_t conn, smtp_session_t session)
   session->rsp_recipient = next_recipient (session->rsp_recipient);
   if (session->rsp_recipient != NULL)
     session->rsp_state = S_rcpt;
-  else
-#ifdef USE_REQUIRE_ALL_RECIPIENTS
-  /* Ugly hack:  if requiring all recpients to succeed make the decision to
-  		 proceed now.  Using this option may make it impossible
-  		 to send mail at all if some recipients are experiencing
-  		 transient errors.  */
-  if (session->require_all_recipients
-      && session->current_message->failed_recipients > 0)
+  else if (session->require_all_recipients
+           && session->current_message->failed_recipients > 0)
     {
       reset_status (&session->current_message->message_status);
       session->rsp_state = next_message (session) ? S_rset : S_quit;
     }
   else
-#endif
-#if 0
+#ifdef USE_CHUNKING
     session->rsp_state = (session->extensions & EXT_CHUNKING) ? S_bdat : S_data;
 #else
     session->rsp_state = S_data;
@@ -1166,7 +1200,7 @@ rsp_rcpt (siobuf_t conn, smtp_session_t session)
 void
 cmd_data (siobuf_t conn, smtp_session_t session)
 {
-  sio_set_timeout (conn, 2 * 60 * 1000);
+  sio_set_timeout (conn, session->data_timeout);
   sio_write (conn, "DATA\r\n", -1);
   session->cmd_state = -1;
 }
@@ -1179,6 +1213,11 @@ rsp_data (siobuf_t conn, smtp_session_t session)
 
   message = session->current_message;
   code = read_smtp_response (conn, session, &message->message_status, NULL);
+  if (code < 0)
+    {
+      session->rsp_state = S_quit;
+      return;
+    }
   if (code == 4 || code == 5)
     {
       /* The server will not accept this message.
@@ -1229,7 +1268,7 @@ cmd_data2 (siobuf_t conn, smtp_session_t session)
       return;
     }
 
-  sio_set_timeout (conn, 3 * 60 * 1000);
+  sio_set_timeout (conn, session->transfer_timeout);
 
   /* Arrange to read the current message from the application. */
   msg_source_set_cb (session->msg_source,
@@ -1282,7 +1321,7 @@ cmd_data2 (siobuf_t conn, smtp_session_t session)
          deleted by the library.  If header == line it is passed
          unchanged otherwise, header must be freed after use. */
       header = process_header (session->current_message, line, &len);
-      if (header != NULL)
+      if (header != NULL && len > 0)
 	{
 	  /* Notify byte count to the application. */
 	  if (session->event_cb != NULL)
@@ -1337,30 +1376,31 @@ break_2:
      application might just send a CRLF followed by the message body.
      Libesmtp will then provide all the necessary headers. */
   while ((header = missing_header (session->current_message, &len)) != NULL)
-    {
-      /* Notify byte count to the application. */
-      if (session->event_cb != NULL)
-	(*session->event_cb) (session, SMTP_EV_MESSAGEDATA,
-	                      session->event_cb_arg,
-	                      session->current_message, len);
+    if (len > 0)
+      {
+	/* Notify byte count to the application. */
+	if (session->event_cb != NULL)
+	  (*session->event_cb) (session, SMTP_EV_MESSAGEDATA,
+				session->event_cb_arg,
+				session->current_message, len);
 
-      if (session->monitor_cb && session->monitor_cb_headers)
-	(*session->monitor_cb) (header, len, SMTP_CB_HEADERS,
-				session->monitor_cb_arg);
-      for (pline = header; pline < header + len; pline = p)
-	{
-	  p = memchr (pline, '\n', header + len - pline);
-	  if (p == NULL)
-	    {
-	      set_errno (ERANGE);
-	      session->cmd_state = session->rsp_state = -1;
-	      return;
-	    }
-	  if (pline[0] == '.')
-	    sio_write (conn, ".", 1);
-	  sio_write (conn, pline, ++p - pline);
-	}
-    }
+	if (session->monitor_cb && session->monitor_cb_headers)
+	  (*session->monitor_cb) (header, len, SMTP_CB_HEADERS,
+				  session->monitor_cb_arg);
+	for (pline = header; pline < header + len; pline = p)
+	  {
+	    p = memchr (pline, '\n', header + len - pline);
+	    if (p == NULL)
+	      {
+		set_errno (ERANGE);
+		session->cmd_state = session->rsp_state = -1;
+		return;
+	      }
+	    if (pline[0] == '.')
+	      sio_write (conn, ".", 1);
+	    sio_write (conn, pline, ++p - pline);
+	  }
+      }
 
   /* ... and finally terminate the message headers */
   sio_write (conn, "\r\n", 2);
@@ -1394,7 +1434,7 @@ break_2:
   sio_write (conn, ".\r\n", 3);
   sio_flush (conn);
 
-  sio_set_timeout (conn, 10 * 60 * 1000);
+  sio_set_timeout (conn, session->data2_timeout);
   session->cmd_state = -1;
 }
 
@@ -1411,6 +1451,11 @@ rsp_data2 (siobuf_t conn, smtp_session_t session)
   code = read_smtp_response (conn, session,
 			     &session->current_message->message_status,
 			     NULL);
+  if (code < 0)
+    {
+      session->rsp_state = S_quit;
+      return;
+    }
 
   if (code == 2)
     {
